bfgs.lm.las <- function(x0, XtX, Xty, lambda, eps) {
  res <- optim(x0, method="L-BFGS-B",
               fn     = L.lasso.gaussian,
               gr     = dL.lasso.gaussian,
               XtX    = XtX,
               Xty    = Xty,
               lambda = lambda,
               control= list(pgtol=eps,factr=eps))
  return(list(i=res$counts[1],xk=res$par))
}

bfgs.lm.grp <- function(x0, XtX, Xty, lambda, pk, eps) {
  res <- optim(x0, method="L-BFGS-B",
               fn     = L.group.gaussian,
               gr     = dL.group.gaussian,
               XtX    = XtX,
               Xty    = Xty,
               lambda = lambda,
               pk     = pk,
               control= list(pgtol=eps,factr=eps))
  return(list(i=res$counts[1],xk=res$par))
}

bfgs.lm.coo <- function(x0, XtX, Xty, lambda, pk, eps, lower, upper) {
  res <- optim(x0, method="L-BFGS-B",
               fn     = L.coop.gaussian,
               gr     = dL.coop.gaussian,
               XtX    = XtX,
               Xty    = Xty,
               lambda = lambda,
               pk     = pk,
               lower = lower,
               upper = upper,
               control= list(factr=eps,pgtol=eps))
  return(list(i=res$counts[1],xk=res$par))
}

ista.lm.las <- function(x0, XtX, Xty, lambda, eps) {
  return(.C("ISTA_LM_LAS",
            as.integer(length(x0)),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

ista.lm.grp <- function(x0, XtX, Xty, lambda, pk, eps) {
  return(.C("ISTA_LM_GRP",
            as.integer(length(x0)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

ista.lm.coo <- function(x0, XtX, Xty, lambda, pk, eps, lower, upper) {
  return(.C("ISTA_LM_COO",
            as.integer(length(x0)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

fista.lm.las <- function(x0, XtX, Xty, lambda, eps) {
  return(.C("FISTA_LM_LAS",
            as.integer(length(x0)),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

fista.lm.grp <- function(x0, XtX, Xty, lambda, pk, eps) {
  return(.C("FISTA_LM_GRP",
            as.integer(length(x0)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

fista.lm.coo <- function(x0, XtX, Xty, lambda, pk, eps, lower, upper) {
  return(.C("FISTA_LM_COO",
            as.integer(length(x0)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(XtX),
            as.double(Xty),
            as.double(lambda),
            as.double(max(eigen(XtX)$values,TRUE,TRUE)),
            as.integer(10000), # max # of iteration
            as.double(eps),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0)))
}

bfgs.lrm.las <- function(x0, X, y, lambda, eps, L0) {
  res <- optim(as.matrix(x0),
               method = "L-BFGS-B",
               fn     = L.lasso.binomial,
               gr     = dL.lasso.binomial,
               X      = X,
               y      = y,
               lambda = lambda,
               control= list(factr=eps,pgtol=eps)) 
  res$par[abs(res$par) < eps] <- 0
  return(list(i=res$counts[1],xk=res$par))
}

bfgs.lrm.grp <- function(x0, X, y, lambda, pk, eps, L0) {
  res <- optim(as.matrix(x0),
               method = "L-BFGS-B",
               fn     = L.group.binomial,
               gr     = dL.group.binomial,
               X      = X,
               y      = y,
               lambda = lambda,
               pk     = pk,
               control= list(factr=eps,pgtol=eps))
  res$par[abs(res$par) < eps] <- 0
  return(list(i=res$counts[1],xk=res$par))
}

bfgs.lrm.coo <- function(x0, X, y, lambda, pk, eps, lower, upper, L0) {
  res <- optim(as.matrix(x0),
               method = "L-BFGS-B",
               fn     = L.coop.binomial,
               gr     = dL.coop.binomial,
               X      = X,
               y      = y,
               lambda = lambda,
               pk     = pk,
               lower = lower,
               upper = upper,
               control= list(factr=eps,pgtol=eps))
  res$par[abs(res$par) < eps] <- 0
  return(list(i=res$counts[1],xk=res$par))
}

ista.lrm.las <- function(x0, X, y, lambda, eps, L0) {
  out <- .C("ISTA_LRM_LAS",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  out$xk[abs(out$xk) < eps] <- 0
  return(out)
}

ista.lrm.grp <- function(x0, X, y, lambda, pk, eps, L0) {
  out <- .C("ISTA_LRM_GRP",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  return(out)
}

ista.lrm.coo <- function(x0, X, y, lambda, pk, eps, lower, upper, L0) {
  out <- .C("ISTA_LRM_COO",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  return(out)
}

fista.lrm.las <- function(x0, X, y, lambda, eps, L0) {
  out <- .C("FISTA_LRM_LAS",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  return(out)
}

fista.lrm.grp <- function(x0, X, y, lambda, pk, eps, L0) {
  out <- .C("FISTA_LRM_GRP",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  return(out)
}

fista.lrm.coo <- function(x0, X, y, lambda, pk, eps, lower, upper, L0) {
  out <- .C("FISTA_LRM_COO",
            as.integer(length(x0)),
            as.integer(length(y)),
            as.integer(length(pk)),
            as.integer(pk),
            as.double(X),
            as.double(y),
            as.double(lambda),
            L = as.double(L0),
            as.integer(10000),
            as.double(eps/length(x0)),
            x0 = as.double(x0),
            xk = as.double(x0),
            i  = as.integer(0))
  return(out)
}

